
#include <X11/Xlib.h>
#include <string.h>
#include <time.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <fcntl.h>

#include "epda.h"

#define REFRESH 200000 	//time interval in usec to refresh display window

struct shared_data *board;
void *shared_memory; 
int shmid;	//shared memory id in integer
int semid;	// semaphore id in integer
struct sembuf semapho;	// struct type for use semaphore.
struct timeval t1, t2, t3, tmpt;


long long  calclock(); 
void task();
void enteringCriticalSection(int num_task);
void leavingCriticalSection(int num_task);
int processCompare(int number_j, int process_j, int number_i, int process_i);
int enteringCriticalSectionSemaphore(void);
int leavingCriticalSectionSemaphore(void);


int main(int argc, char **argv) {
	//Some Definitions for X-window
    Display *d ;
    Window   w, root ;
	Font f;
	GC gc;
	XSetWindowAttributes xswa;

	char msg0[100], msg1[100], msg2[100], msg3[100], mbf1[100], mbf2[100];

	struct   timeval  start_time, pre_time, cur_time, end_time;
	long long elapsedtime;
	int ctime1, ctime2,i;	
	char cmd[100];
	double stime,a;
	FILE *fo;


	//Set envirionments for X-window
    d = XOpenDisplay(NULL) ;
    root = XDefaultRootWindow (d);
    w = XCreateSimpleWindow ( d, root, 260, 110, 400, 100, 2, BlackPixel (d,0), WhitePixel(d,0) );
	XChangeWindowAttributes (d, w, CWOverrideRedirect, &xswa);
    XMapWindow (d, w);
	gc = XCreateGC(d,w,0L, (XGCValues *) NULL);
	f = XLoadFont (d,"fixed");
	XSetFont (d,gc,f);

	//set the starting time
	gettimeofday(&start_time,NULL);
	pre_time = start_time;
	
	// open log file
	if((flog = fopen("call.log", "w")) == NULL) {
		perror("file open error");
		exit(-1);
	}

	// get the id of the message queue created by main task
	mqid = atoi(argv[1]);
	//printf("Call mqid : %d\n", mqid);

	shmid = atoi(argv[6]);	// get the shared memory id in string from argv[]
	shared_memory = shmat(shmid, 0, 0);	// attach the shared memory to its own memory space
	if(shared_memory == (void *)-1) {
		perror("shmat failed:");
		exit(1);
	}
	board = (struct shared_data *)shared_memory;		//mount the attached shared memory to the structure of board 
		
	semid = atoi(argv[7]);	// Get the semaohore ID
	


	while(1) {
		for(i = 0; i < 100; i++)
			msg0[i] = msg1[i] = msg2[i] = msg3[i] = mbf1[i] = mbf2[i] = '\0';	// initizlize message buffers
		sprintf(msg0,"call task: ");
		
		task();
		
		gettimeofday(&cur_time,NULL);
		elapsedtime = calclock(pre_time, cur_time);

		if(elapsedtime > REFRESH) { //update display window every REFRESH us
			pre_time = cur_time;	//replace previous time with current time

			//**************************************************************************************
			// obtain accumulated CPU time which is the sum of the 42 and 43'th fields of /proc/ppid/stat
			//**************************************************************************************
			// input the result of "cat /proc/ppid/stat" to "awk '{printf $42}'"
			sprintf(cmd, "cat /proc/%d/stat |awk '{printf $42}'", getpid());
			// input the result of "awk '{printf $42}'",i.e., the 42'th field, to ctime1 by means of pipe
			fo = popen(cmd, "r");
			fscanf(fo, "%d", &ctime1);
			fclose(fo);
			// input the result of "cat /proc/ppid/stat" to "awk '{printf $43}'"
			sprintf(cmd, "cat /proc/%d/stat |awk '{printf $43}'", getpid());
			// input the result of "awk '{printf $42}'",i.e., the 42'th field, to ctime1 by means of pipe
			fo = popen(cmd, "r");
			fscanf(fo, "%d", &ctime2);
			fclose(fo);
			// obtain accumulated CPU time through adding two fields and normalizing it in terms of second unit
			a = ctime1;
			a += ctime2;
			a = a / 100;

			// obtain accumulated service time of the differece between the starting and current times, and normalize it in terms of second unit
			stime = calclock(start_time, cur_time);
			stime = stime / 1000000;

			//printf("service time = %1.3f [sec], CPU time =%1.3f [Sec]\n", stime,a);

			// make three messages to display on the window
			sprintf(mbf1, "pid : %d | ppid : %d,", getpid(), getppid());
			sprintf(mbf2, " %1.1f/%1.1f", a,stime);
			strcat(msg0, mbf1);
			strcat(msg0, mbf2);
			sprintf(msg1, "This is Call task!");
			sprintf(msg2, "Insert numbers..");
			sprintf(msg3, "Charge = ");

			// After clear the window, display four messages on the window
			XClearWindow(d,w);
			XDrawString(d,w,gc,5,20,msg0,sizeof(msg0));
			XDrawString(d,w,gc,15,45,msg1,sizeof(msg1));
			XDrawString(d,w,gc,15,65,msg2,sizeof(msg2));
			XDrawString(d,w,gc,15,85,msg3,sizeof(msg3));
        	XFlush (d);

		}
	}

	//close procedure for X-window
	XUnloadFont (d,f);
	XFreeGC(d,gc);
	XDestroyWindow(d,w);
    XCloseDisplay (d);
 }

void task()
{
	struct MessageFormat rmessage;
	struct MSMSTextFormat *rmtext;
	struct tm t;
	char buf[100], rfile_name[50], wfile_name[50], insert_tlog[100];
	FILE *wfp, *rfp;
	int readCnt, writeCnt, open_file, write_file;
	int open_log_file, i, check_count, count, tmp_count;
	double take_time, tmp_time;

	// there is no message to recieve, return emmediately
	if(msgrcv(mqid, (void *)&rmessage, BUFSIZ, CALL, IPC_NOWAIT) == -1)
		return;
	
	MessageLog(flog, "RX", rmessage);
	fprintf(stderr, "Call Test\n");
	nrmessage++;
	
	rmtext = (struct MSMSTextFormat *)rmessage.text;
	
	check_count = rmtext->NumberOfMessages;
	strcpy(rfile_name, rmtext->data);
	strcpy(wfile_name, rmtext->RecipientNumber);
	fprintf(stderr, "Send message to %s. \n", rmtext->RecipientNumber);
		
	for(count = 0; count < check_count; count++) {	

		gettimeofday(&t1, NULL);

						
/*		gettimeofday(&t2, 0);
		t3.tv_sec = t2.tv_sec - t1.tv_sec;
		t3.tv_usec = t2.tv_usec - t1.tv_usec;
		if(t3.tv_usec < 0) {
			t3.tv_sec --;
			t3.tv_usec += 1000000;
		}

		sprintf(insert_tlog, "%ld.%ld\t%ld.%ld\n", t2.tv_sec, t2.tv_usec, t3.tv_sec, t3.tv_usec); 
	
		rfp = fopen(rfile_name, "r");
		wfp = fopen(wfile_name, "a");
		while(buf != 0) {
			fscanf(rfp, "%s", &buf);
			fprintf(wfp, "%s", buf);
		}
		fclose(rfp);
		fclose(wfp);
	*/
		//fprintf(stderr, "open read file name : %s\n", rfile_name);
		//fprintf(stderr, "open write file name : %s\n", wfile_name);
		open_file = open(rfile_name, O_RDONLY);
		write_file = open(wfile_name, O_WRONLY | O_CREAT | O_APPEND);
		if(open_file < 0 || write_file < 0) {
			fprintf(stderr, "File open failed - call\n");
			exit(-1);
		}
		for(readCnt = 1; readCnt > 0;) {
			readCnt = read(open_file, buf, 100);
			writeCnt = write(write_file, buf, readCnt);
		}
		close(open_file);
		close(write_file);

		gettimeofday(&t2, 0);
		t3.tv_sec = t2.tv_sec - t1.tv_sec;
		t3.tv_usec = t2.tv_usec - t1.tv_usec;
		if(t3.tv_usec < 0) {
			t3.tv_sec --;
			t3.tv_usec += 1000000;
		}
		
		take_time = t3.tv_sec + (t3.tv_usec * 0.0000001);

		//enteringCriticalSection(CALL);		// Enter the Critical Section.
		if(!leavingCriticalSectionSemaphore())	// Leave the Critical Section. But if return value is error, display error message.
			exit(-1);
	
		for(i = 0; i < 100000000; )
			i++;

		board->ase[TASKS - 1]++;
		board->ast[TASKS - 1].tv_sec += t3.tv_sec;
		board->ast[TASKS - 1].tv_usec += t3.tv_usec;
		if(board->ast[TASKS - 1].tv_usec > 1000000) {
			board->ast[TASKS - 1].tv_usec -= 1000000;
			board->ast[TASKS - 1].tv_sec++;
		}
		
		board->ase[WEB]++;
		board->ast[WEB].tv_sec += t3.tv_sec;
		board->ast[WEB].tv_usec += t3.tv_usec;
		if(board->ast[CALL].tv_usec > 1000000) {
			board->ast[CALL].tv_usec -= 1000000;
			board->ast[CALL].tv_sec++;
		}		
		
		//leavingCriticalSection(CALL);		// Leave the Critical Section.
		if(!leavingCriticalSectionSemaphore())	// Leave the Critical Section. But if return value is error, display error message.
			exit(-1);
	}
	//write(Pipe2Main, "complete", BUFSIZ);
	fprintf(stderr, "Mail send complete!\n");
	return;
}	

long long  calclock( struct timeval clock_a, struct timeval clock_b ) 
{
	long long timedelay, temp, temp_n;
	if (clock_a.tv_usec >= clock_a.tv_usec) {
		temp = clock_b.tv_sec - clock_a.tv_sec;
		temp_n = clock_b.tv_usec - clock_a.tv_usec;
		timedelay = 1000000 * temp + temp_n;
	} else {
		temp = clock_b.tv_sec - clock_a.tv_sec - 1;
		temp_n = 1000000 + clock_b.tv_usec - clock_a.tv_usec;
		timedelay = 1000000 * temp + temp_n;
	}
	return timedelay;
}

int enteringCriticalSectionSemaphore(void)
{
	semapho.sem_num = 0;
	semapho.sem_op = -1;
	semapho.sem_flg = SEM_UNDO;
	if(semop(semid, &semapho, 1) == -1) {
		fprintf(stderr, "failed to entering critical section by semaphore!\n");
		return 0;
	}
	return 1;
}

int leavingCriticalSectionSemaphore(void)
{
	semapho.sem_num = 0;
	semapho.sem_op = 1;
	semapho.sem_flg = SEM_UNDO;
	if(semop(semid, &semapho, 1) == -1) {
		fprintf(stderr, "failed to leaving critical section by semaphore!\n");
		return 0;
	}
	return 1;
}

void enteringCriticalSection(int num_task)
{
	int i, max_num;

	max_num = 0;
	board->choosing[num_task] = TRUE;
	for(i = 0; i < TASKS; i++) {
		if(max_num < board->number[i])
			max_num = board->number[i];
	}
	board->number[num_task] = max_num + 1;
	board->choosing[num_task] = FALSE;

	for(i = 0; i < TASKS; i++) {
		while(board->choosing[i])
			;
		while(board->number[i] && processCompare(board->number[i], i, board->number[num_task], num_task))
			;
	}
}

void leavingCriticalSection(int num_task)
{
	board->number[num_task] = 0;
}

int processCompare(int number_j, int process_j, int number_i, int process_i)
{
	if(number_j < number_i)
		return TRUE;
	else if(process_j < process_i)
		return TRUE;
	else
		return FALSE;
}
