
#include <X11/Xlib.h>
#include <string.h>
#include <time.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <fcntl.h>
#include <sys/sem.h>

#include "epda.h"

#define REFRESH 200000 	// time interval in usec to refresh display window

struct shared_data *board;
void *shared_memory; 
int shmid;	//shared memory id in integer
int semid;	// semaphore id in integer
struct timeval t1, t2, t3, tmpt;
int check_count, count;		// check send messages
struct sembuf semapho;	// struct type for use semaphore.

void task(int Pipe2Main);
long long  calclock(); 
void enteringCriticalSection(int num_task);
void leavingCriticalSection(int num_task);
int processCompare(int number_j, int process_j, int number_i, int process_i);
int leavingCriticalSectionSemaphore(void);		// entering critical section by semaphore.
int enteringCriticalSectionSemaphore(void);	// leaving critical section by semaphore.

 

int main(int argc, char **argv) {
	//Some Definitions for X-window
    Display *d ;
    Window   w, root ;
	Font f;
	GC gc;
	XSetWindowAttributes xswa;

	char msg0[100], msg1[100], msg2[100], msg3[100], mbf1[100], mbf2[100];
	struct   timeval  start_time, pre_time, cur_time, end_time;
	long long elapsedtime;
	int ctime1, ctime2,i;	
	char cmd[100];
	double stime, a;
	FILE *fo;
	int Pipe2Main[2], Pipe2Me[2];


	// Set envirionments for X-window
    d = XOpenDisplay(NULL) ;
    root = XDefaultRootWindow (d);
    w = XCreateSimpleWindow(d, root, 260, 110, 400, 100, 2, BlackPixel (d, 0), WhitePixel(d, 0));
	XChangeWindowAttributes(d, w, CWOverrideRedirect, &xswa);
    XMapWindow(d, w);
	gc = XCreateGC(d, w, 0L, (XGCValues *)NULL);
	f = XLoadFont(d, "fixed");
	XSetFont(d, gc, f);

	// set the starting time
	gettimeofday(&start_time, NULL);
	pre_time = start_time;
	
	// open log file
	if((flog = fopen("sms.log", "w")) == NULL) {
		perror("file open error");
		exit(-1);
	}

	// get the id of the message queue created by main task
	mqid = atoi(argv[1]);

	// get pipe IDs and make two one way pipes.
	Pipe2Main[0] = atoi(argv[2]);
	Pipe2Main[1] = atoi(argv[3]);
	Pipe2Me[0] = atoi(argv[4]);
	Pipe2Me[1] = atoi(argv[5]);
	close(Pipe2Main[0]);
	close(Pipe2Me[1]);
	//fprintf(stderr, "Pipe2Main[1] = %d, Pipe2Me[0] = %d", Pipe2Main[1], Pipe2Me[0]);

	shmid = atoi(argv[6]);	//get the shared memory id in string from argv[]
	shared_memory = shmat(shmid,0,0);	//attach the shared memory to its own memory space
	if(shared_memory == (void *) -1) {
		perror("shmat failed:");
		exit(1);
	}
	board = (struct shared_data *)shared_memory;		//mount the attached shared memory to the structure of board 
		
	// get the id of the semaphore created by main task
	semid = atoi(argv[7]);

	while(1) {
		for(i = 0; i < 100; i++)
			msg0[i] = msg1[i] = msg2[i] = msg3[i] = mbf1[i] = mbf2[i] = '\0';	// initizlize message buffers
		sprintf(msg0, "SMS task: ");
		
		task(Pipe2Main[1]);
		
		
		gettimeofday(&cur_time, NULL);
		elapsedtime = calclock(pre_time, cur_time);

		if(elapsedtime > REFRESH) {	// update display window every REFRESH us
			pre_time = cur_time;	// replace previous time with current time

			//**************************************************************************************
			// obtain accumulated CPU time which is the sum of the 42 and 43'th fields of /proc/ppid/stat
			//**************************************************************************************
			// input the result of "cat /proc/ppid/stat" to "awk '{printf $42}'"
			sprintf(cmd, "cat /proc/%d/stat |awk '{printf $42}'", getpid());
			// input the result of "awk '{printf $42}'",i.e., the 42'th field, to ctime1 by means of pipe
			fo = popen(cmd, "r");
			fscanf(fo, "%d", &ctime1);
			fclose(fo);
			// input the result of "cat /proc/ppid/stat" to "awk '{printf $43}'"
			sprintf(cmd, "cat /proc/%d/stat |awk '{printf $43}'", getpid());
			// input the result of "awk '{printf $42}'",i.e., the 42'th field, to ctime1 by means of pipe
			fo = popen(cmd, "r");
			fscanf(fo, "%d", &ctime2);
			fclose(fo);
			// obtain accumulated CPU time through adding two fields and normalizing it in terms of second unit
			a = ctime1;
			a += ctime2;
			a = a / 100;

			// obtain accumulated service time of the differece between the starting and current times, and normalize it in terms of second unit
			stime = calclock(start_time, cur_time);
			stime = stime / 1000000;

			// printf("service time = %1.3f [sec], CPU time =%1.3f [Sec]\n", stime,a);

			// make three messages to display on the window
			sprintf(mbf1, "pid : %d | ppid : %d,", getpid(), getppid());
			sprintf(mbf2, " %1.1f/%1.1f", a, stime);
			strcat(msg0, mbf1);
			strcat(msg0, mbf2);
			sprintf(msg1, "This is SMS task!");
			sprintf(msg2, "NumberOfMessages : %d/%d", count, check_count);
			sprintf(msg3, "Charge = ");

			// After clear the window, display four messages on the window
			XClearWindow(d, w);
			XDrawString(d, w, gc, 5, 20, msg0, sizeof(msg0));
			XDrawString(d, w, gc, 15, 45, msg1, sizeof(msg1));
			XDrawString(d, w, gc, 15, 65, msg2, sizeof(msg2));
			XDrawString(d, w, gc, 15, 85, msg3, sizeof(msg3));
        	XFlush(d);
		}
	}

	//close procedure for X-window
	XUnloadFont (d,f);
	XFreeGC(d,gc);
	XDestroyWindow(d,w);
    XCloseDisplay (d);
 }

// task module
void task(int Pipe2Main)
{
	struct MessageFormat rmessage;
	struct MSMSTextFormat *rmtext;
	struct SMSTextFormat *rstext;
	struct PBookTextFormat *rptext;
	struct MessageFormat smessage;
	struct MSMSTextFormat smtext;
	struct SMSTextFormat sstext;
	struct PBookTextFormat sptext;
	struct tm t;
	char buf[100], file_name[20], insert_tlog[100];
	FILE *fp;
	int open_log_file, i, tmp_count, flag;
	struct timeval tmp_time;

	// there is no message to recieve, return emmediately
	if(msgrcv(mqid, (void *)&rmessage, BUFSIZ, SMS, IPC_NOWAIT) == -1)
		return;

	MessageLog(flog, "RX", rmessage);
	nrmessage++;
	
	rmtext = (struct MSMSTextFormat *)rmessage.text;
	
	if(rmtext->SourceTaskNumber == MAIN && rmtext->code == MULTIPLESMSREQUEST) {
		check_count = rmtext->NumberOfMessages;
		strcpy(file_name, rmtext->ServiceTimeLogFileName);
		fprintf(stderr, "time check log file : %s\n", file_name);
	
		smessage.type = PBOOK;
		sptext.SourceTaskNumber = SMS;
		sptext.code = REQUEST;
		strcpy(sptext.RequestName, rmtext->RecipientName);
		strcpy(sptext.data, rmtext->data);
		StructCopy(smessage.text, &sptext, BUFSIZ);
	
		for(count = 0; count < check_count; count++) {		
		
			gettimeofday(&t1, 0);
		
			// send message to PBOOK
			if(msgsnd(mqid, (void *)&smessage, BUFSIZ, IPC_NOWAIT) == -1) {
				fprintf(stderr, "Fail to send message!!");
				exit(-1);
			}
			MessageLog(flog, "TX", smessage);
			nsmessage++;
		
			// receive message from PBOOK
			while(msgrcv(mqid, (void *)&rmessage, BUFSIZ, SMS, IPC_NOWAIT) == -1) {
			//	fprintf(stderr, "Fail to receive message from PBOOK!!\n");
			//	exit(-1);
				//return;
			}
			MessageLog(flog, "RX", rmessage);
			nrmessage++;
			rptext = (struct PBookTextFormat *)rmessage.text;
		
			gettimeofday(&t2, 0);
			t3.tv_sec = t2.tv_sec - t1.tv_sec;
			t3.tv_usec = t2.tv_usec - t1.tv_usec;
			if(t3.tv_usec < 0) {
				t3.tv_sec --;
				t3.tv_usec += 1000000;
			}
			
			fp = fopen(file_name, "a");
			fprintf(fp, "%ld.%ld \t %ld.%ld\n", t2.tv_sec, t2.tv_usec, t3.tv_sec, t3.tv_usec);
			fclose(fp);
			//open_log_file = open(file_name, O_RDWR | O_APPEND | O_CREAT);
			//write(open_log_file, insert_tlog, i);
			//close(open_log_file);
			smessage.type = PBOOK;
			sptext.SourceTaskNumber = SMS;
			sptext.code = REQUEST;
			strcpy(sptext.RequestName, rptext->RequestName);
			strcpy(sptext.data, rptext->data);
			StructCopy(smessage.text, &sptext, BUFSIZ);
				
			//enteringCriticalSection(SMS);		// Enter the Critical Section.
			//if(!enteringCriticalSectionSemaphore())
				//exit(-1);
			
			tmp_count = board->ase[TASKS - 1];
			tmp_time.tv_sec = board->ast[TASKS - 1].tv_sec;
			tmp_time.tv_usec = board->ast[TASKS - 1].tv_usec;
			
			for(i = 0; i < 1000000; )
				i++;
			tmp_count++;
			board->ase[TASKS - 1] = tmp_count;
			board->ast[TASKS - 1].tv_sec = tmp_time.tv_sec + t3.tv_sec;
			board->ast[TASKS - 1].tv_usec = tmp_time.tv_usec + t3.tv_usec;
			if(board->ast[TASKS - 1].tv_usec > 1000000) {
				board->ast[TASKS - 1].tv_usec -= 1000000;
				board->ast[TASKS - 1].tv_sec++;
			}
		
			board->ase[SMS]++;
			board->ast[SMS].tv_sec += t3.tv_sec;
			board->ast[SMS].tv_usec += t3.tv_usec;
			if(board->ast[SMS].tv_usec > 1000000) {
				board->ast[SMS].tv_usec -= 1000000;
				board->ast[SMS].tv_sec++;
			}
			
			//leavingCriticalSection(SMS);		// Leave the Critical Section.
			//if(!leavingCriticalSectionSemaphore())
				//exit(-1);



		}
		//write(Pipe2Main, "complete", BUFSIZ);
		fprintf(stderr, "MSMS complete!\n");
	}

	return;
}

int enteringCriticalSectionSemaphore(void)
{
	semapho.sem_num = 0;
	semapho.sem_op = -1;
	semapho.sem_flg = SEM_UNDO;
	if(semop(semid, &semapho, 1) == -1) {
		fprintf(stderr, "failed to entering critical section by semaphore!\n");
		return 0;
	}
	return 1;
}

int leavingCriticalSectionSemaphore(void)
{
	semapho.sem_num = 0;
	semapho.sem_op = 1;
	semapho.sem_flg = SEM_UNDO;
	if(semop(semid, &semapho, 1) == -1) {
		fprintf(stderr, "failed to leaving critical section by semaphore!\n");
		return 0;
	}
	return 1;
}


void enteringCriticalSection(int num_task)
{
	int i, max_num;

	max_num = 0;
	board->choosing[num_task] = TRUE;
	for(i = 0; i < TASKS; i++) {
		if(max_num < board->number[i])
			max_num = board->number[i];
	}
	board->number[num_task] = max_num + 1;
	board->choosing[num_task] = FALSE;

	for(i = 0; i < TASKS; i++) {
		while(board->choosing[i])
			;
		while(board->number[i] && processCompare(board->number[i], i, board->number[num_task], num_task))
			;
	}
}

void leavingCriticalSection(int num_task)
{
	board->number[num_task] = 0;
}

int processCompare(int number_j, int process_j, int number_i, int process_i)
{
	if(number_j < number_i)
		return TRUE;
	else if(process_j < process_i)
		return TRUE;
	else
		return FALSE;
}

long long  calclock( struct timeval clock_a, struct timeval clock_b ) 
{
	long long timedelay, temp, temp_n;
	if (clock_a.tv_usec >= clock_a.tv_usec) {
		temp = clock_b.tv_sec - clock_a.tv_sec;
		temp_n = clock_b.tv_usec - clock_a.tv_usec;
		timedelay = 1000000 * temp + temp_n;
	} else {
		temp = clock_b.tv_sec - clock_a.tv_sec - 1;
		temp_n = 1000000 + clock_b.tv_usec - clock_a.tv_usec;
		timedelay = 1000000 * temp + temp_n;
	}
	return timedelay;
}
