/***************************************************************************
 *            main.c
 *
 *  Fri Oct 19 01:06:52 2007
 *  Copyright  2007  pchero
 *  pchero21@gmail.com
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 

#include <string.h>
#include <time.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <sys/sem.h>
#include <termios.h>
//#include <term.h>
//#include <curses.h>


static struct termios initial_settings, new_settings;

#include "epda.h"

#define TASKS 11

//Each record of task table has two fields of process id and task name
struct tasktable {
	int ProcessId;
	char TaskName[50];
	int Pipe2Task;	//pipe to child process
	int Pipe2Main;	//pipe from child process
};

struct tasktable tb[TASKS];
struct shared_data *board;
void *shared_memory; 
key_t shmkey;			// key to create a shared memory
int shmid;					// shared memory id in integer
int semid;					// semaphore id in integer
char cshmid[10]; 		// shared memory id in character, which is to hand over the shared memory id to a child process via argv[]
char csemid[10];		// semaphore id in character, which is to hand over the semaphore id to a child process via argv[]
struct sembuf semapho;	// struct type for use semaphore.

void init_keyboard(); 
void close_keyboard(); 
int kbhit(); 
int readch(); //for asynchronous(unblock) MMC
void sms();
void call();
void mail();
void statistics();

int main()
{
	// Define the task table tb[i] of size TASKS, where i is task number
	int button, i;
	char c[10], cmqid[20], pipeid[4][50];
	key_t mqkey;	// define key to identify message queue. Notice that the key is a kind of message queue name
	key_t smkey;	// define key to identify semaphore. Notice that the key is a kind of semaphore name
	int Pipe2Main[TASKS][2], Pipe2Task[TASKS][2];
	struct msqid_ds *buf;

	if((flog = fopen("main.log","w")) == NULL) {
		fprintf(stderr, "Fail to oepn main.log file");
		exit(-1);
	}
	if((mqkey = ftok(".",'m')) == -1) {	// obtain a key needed to create a message queue. The key is similar to file name needed to create a file
		if((mqkey = ftok(".",'e')) == -1) {
			fprintf(stderr, "Fail to obtain mkey !!");
			exit(-1);
		}
	}
	// create a message queue with key for message communication among TASKS tasks in ePDA which is handled by message queue ID (mqid)
	// If alreay message queue, recreate message queue after remove the current message queue which is to clear unexpected garbage up 
	if((msgget(mqkey, 0666|IPC_CREAT|IPC_EXCL)) == -1) {
		    msgctl(mqid, IPC_RMID, buf); // remove the message queue of mqid
		if((mqid = msgget(mqkey, 0666|IPC_CREAT)) == -1) {
			fprintf(stderr, "Fail to obtain mqid at main!!");
			exit(-1);
		}
	}

	// Initialize pipe channels between parent and all child processes 
	for(i = 0; i < TASKS; i++) {
		if(pipe(Pipe2Main[i]) < 0) 
			fprintf(stderr, "pipe open error");
		if(pipe(Pipe2Task[i]) < 0) 
			fprintf(stderr, "pipe open error");;
//		close(Pipe2Main[i][1]);close(Pipe2Task[i][0]);//make a pipe channel
		tb[i].Pipe2Task = Pipe2Task[i][1];	// PipeOut[i][1] : pipe to task i	
		tb[i].Pipe2Main = Pipe2Main[i][0];	// PipeIn[i][0] : pipe from task i 
	}
		
	sprintf(cmqid, "%d", mqid);	// convert to string which is to put mqid in argv[] array
	
	// initialize tasktable
	for(i = 0; i < TASKS; i++)
		tb[i].ProcessId = 0;
	strcpy(tb[CALL].TaskName, "Call");
	strcpy(tb[CONVERSATION].TaskName, "Conversation");
	strcpy(tb[SMS].TaskName, "SMS");
	strcpy(tb[PBOOK].TaskName, "PBook");
	strcpy(tb[MP3].TaskName, "MP3");
	strcpy(tb[DMB].TaskName, "DMB");
	strcpy(tb[CAMERA].TaskName, "Camera");
	strcpy(tb[WEB].TaskName, "Web");
	strcpy(tb[NETWORK].TaskName, "Network");


	if((shmkey = ftok(".", 's')) == -1) {		// obtain an unique key needed to create a shared memory. 
		if((shmkey = ftok(".", 'o')) == -1) {
			fprintf(stderr, "Fail to obtain shmkey !!");
			exit(-1);
		}
	}

	shmid = shmget(shmkey, sizeof(board), 0666 | IPC_CREAT);	// create a shared memory
	if(shmid == -1) {
		perror("shmget failed:");
		exit(1);
	}
	shared_memory = shmat(shmid, 0, 0);	// attach the created shared memory to its own memory space
	if(shared_memory == (void *)-1) {
		perror("shmat failed:");
		exit(1);
	}
	board = (struct shared_data *)shared_memory;		//mount the attached shared memory to the structure of board 
	
	for(i = 0; i < TASKS; i++) {
		board->ase[i] = board->ast[i].tv_sec = board->ast[i].tv_usec = 0.0;		// initialize
		board->choosing[i] = FALSE;
		board->number[i] = 0;
	}

	sprintf(cshmid, "%d", shmid);	// transform the shared memory id in integer into string which is to put the shmid value in argv[] array

	if((smkey = ftok(".",'b')) == -1) {	// obtain a key needed to create a semaphore. The key is similar to file name needed to create a file
		if((smkey = ftok(".",'t')) == -1) {
			fprintf(stderr, "Fail to obtain mkey !!");
			exit(-1);
		}
	}

	semid = semget((key_t)smkey, 1, 0666 | IPC_CREAT);	// Get the semaphore id use semget().
	sprintf(csemid, "%d", semid);

	if(semctl(semid, 0, SETVAL, 1) == -1)							// Initialized semaphore
		fprintf(stderr, "failed initialized semaphore\n");
	
	while(1) {
		//Ask what kinds of action does user want to do
		printf("Button: 0(task ON), 1(task OFF), 2(sms), 3(camera), 4(call), 5(mail), 6(statistics), 9(esc),10(stop)\n");
		scanf("%s",&c);	
		if(c != "s")
			button = atoi(c);	
		else
			fprintf(stderr, "remove some possible garbage occurred when type in 's' to stop a camera service  \n");
		// scanf("%d",&button);	
		switch(button) {
			case 0:
				// read task number that ePDA user pushes ON
				printf("ON: 1(call),2(conversation),3(sms),4(pbook),5(MP3),6(DMB),7(camera),8(web),9(network),10(esc)\n");
				scanf("%d", &button);	
				if(button !=10 && button >=1 && button<=9) {
					// Check whether the corresponding process is alreay running or not
					// Notice that the ProcessId of 0 in task table means there is no running process for the task
					if(tb[button].ProcessId == 0) {
						pid_t pid = fork();	// fork new process
						if(pid == 0) {
							// if the new child process forked just before, prepare parameters such as mqid and pipe IDs to handover the child process in string form 
							fprintf(stderr, "	Success to fork the task of %s\n",tb[button].TaskName); 
							sprintf(pipeid[0], "%d", Pipe2Main[button][0]);
							sprintf(pipeid[1], "%d", Pipe2Main[button][1]);
							sprintf(pipeid[2], "%d", Pipe2Task[button][0]);
							sprintf(pipeid[3], "%d", Pipe2Task[button][1]);
							char *argv[] = {tb[button].TaskName, cmqid, pipeid[0], pipeid[1], pipeid[2], pipeid[3], cshmid, csemid, NULL};//put all parameters to handover to the new process in argv
							//char *argv[] = {tb[button].TaskName, cmqid, pipeid[0], pipeid[1], pipeid[2], pipeid[3], NULL};	// put all parameters to handover to the new process in argv
							// and then execute the corresponding task image file with argc and argv[]
							execv(argv[0], argv);
						} else if(pid < 0) 
							fprintf(stderr, "	Fail to fork the task of %s\n", tb[button].TaskName); 
						else {
							tb[button].ProcessId = pid;	// the return value from fork(), i.e., pid is the id of the new process forked just before
							// make two one-way pipes of Pipe2Main and Pipe2Task 
							close(Pipe2Main[button][1]);
							close(Pipe2Task[button][0]);	// make a pipe channel
						}
					} else 
						fprintf(stderr, "	The task of %s is already running\n",tb[button].TaskName); 
				}
				break;
			
			case 1:
			// read task number that ePDA user pushes OFF
				printf("OFF: 1(call),2(conversation),3(sms),4(pbook),5(MP3),6(DMB),7(camera),8(web),9(network),10(esc)\n");
				scanf("%d", &button);	
				if(button != 10 && button >= 1 && button <= 9) {
					// Check whether the corresponding process is running now or not
					// Notice that the ProcessId of non-zero in task table means there is a running process for the task
					if(tb[button].ProcessId != 0) {
						kill(tb[button].ProcessId, SIGKILL);	// kill the process of the task to be OFF. Notice that the process id can be obtained from the task table in which the process's pid was registered when forked
						tb[button].ProcessId = 0;	// reset the corresponding ProcessId field in the task table
						fprintf(stderr, "	Success to kill the task of %s\n", tb[button].TaskName); 
					} else fprintf(stderr, "	The task of %s is already OFF\n", tb[button].TaskName); 
				}
				break;

			case 2:
				sms();
				break;

			case 3:
				camera();
				break;

			case 4:
				call();
				break;

			case 5:
				mail();
				break;

			case 6:
				statistics();
				break;

			case 7:

			case 8:

			case 9:
				break;

			case 10:	// if stop, kill all child processes and then exit 
				for(i = 0; i < TASKS; i++)
					if(tb[i].ProcessId != 0)
						kill(tb[i].ProcessId, SIGKILL);
				exit(0);
		}
	}
	fclose(flog);
}


//MMC(Man-Machine Communication) routine for the sms task 
void sms() {
	struct MessageFormat smessage;
	struct MessageFormat rmessage;
	struct SMSTextFormat sstext;
	struct SMSTextFormat rstext;
	struct MSMSTextFormat smtext;
	struct MSMSTextFormat *rmtext;
		
	char name[20],mbuf[100], buf[10];
	int multi, count, button, check_end, nc;

	printf("\nButton: 0(send new message),1(exit)\n");
	scanf("%d", &button);	
	if(button == 0) {
		printf("Do you want multilple translation?\n0[yes]/1[no] : ");
		scanf("%d", &multi);
		printf("How many times want to send?\n");
		scanf("%d", &count);
		printf("What is the receiver's name?\n");
		scanf("%s", &name);
		printf("Type the message that you'd like to send (enter enter Key to finish)\n");
		scanf("%s", &mbuf);
			
		if(!multi) {
			// send the MSMS sending request message to the sms task		
			smessage.type = SMS;
			smtext.SourceTaskNumber = MAIN;
			smtext.code = MULTIPLESMSREQUEST;
			strcpy(smtext.RecipientName, name);
			smtext.NumberOfMessages = count;
			strcpy(smtext.ServiceTimeLogFileName, "ServiceTime.log");	// log file for serviceTime
			strcpy(smtext.data, mbuf);
			StructCopy(smessage.text, &smtext, BUFSIZ);
		
			if(msgsnd(mqid, (void *)&smessage, BUFSIZ, IPC_NOWAIT) == -1) {
				fprintf(stderr, "Fail to send message !!");
				exit(-1);
			}
			//fprintf(stderr, "message send to SMS TASK\n");
			MessageLog(flog,"TX",smessage);
			nsmessage++;
		}
		else {
			//send the SMS sending request message to the sms task
			smessage.type = SMS;
			sstext.SourceTaskNumber = MAIN;
			sstext.code = REQUEST;
			strcpy(sstext.RecipientName, name); 
			strcpy(sstext.data, mbuf);
			StructCopy(smessage.text, &sstext, BUFSIZ);
			//Send the prepared message 
			if(msgsnd(mqid, (void *)&smessage, BUFSIZ, IPC_NOWAIT) == -1) {
				fprintf(stderr, "Fail to send message !!");
				exit(-1);
			}
			//fprintf(stderr, "message send to SMS TASK\n");
			MessageLog(flog,"TX",smessage);
			nsmessage++;
		}
	
/*		while(1) {
			nc = read(tb[SMS].Pipe2Main, buf, BUFSIZ);
			if(strcmp(buf, "complete") == 0)
				break;
			//nc = read(tb[SMS].Pipe2Main, buf, BUFSIZ);
			fprintf(stderr, "fofo\n");
		//}
		}
		
		fprintf(stderr, "receive complete message from SMS task!!\n");
		//MessageLog(flog, "RX", rmessage);
		//nrmessage++;
		//rmtext = (struct MSMSTextFormat *)rmessage.text;
		//return;
		*/
	}
}

//MMC(Man-Machine Communication) routine for the camera task 
camera() {
	struct MessageFormat smessage;
	struct MovieTextFormat stext;
	struct MessageFormat rmessage;
	struct MovieTextFormat *rtext;
	char ch, url[50], quality[50], buf[10];
	int button, nc, flag = 0;

	printf("\nButton: 0(movie relay),1(exit)\n");
	scanf("%d", &button);	
	if(button == 0) {
		printf("What is the destination URL?\n");
		scanf("%s", &url);
		printf("What about the quality of the movie ? (resolution/fps)\n");
		scanf("%s", &quality);
		printf("OK ! Press 's' key to stop the movice relay service\n");
			
		// send the movie relay request message to the camera task
		smessage.type = CAMERA;
		stext.SourceTaskNumber = MAIN;		
		stext.code = REQUEST;		
		strcpy(stext.url, url); 
		StructCopy(smessage.text, &stext, BUFSIZ); 
		// Send the prepared message 
		if(msgsnd(mqid, (void *)&smessage, BUFSIZ, IPC_NOWAIT) == -1) {
			fprintf(stderr, "Fail to send message !!");
			exit(-1);
		}
		MessageLog(flog, "TX", smessage);
		nsmessage++;

		// wait the movie relay start message from camera
		if(msgrcv(mqid, (void *)&rmessage, BUFSIZ, MAIN, 0) == -1) {
			fprintf(stderr, "Fail to receive the movie relay start message at main !!");
			exit(-1);
		}
		MessageLog(flog, "RX", rmessage);
		nrmessage++;
		rtext=(struct MovieTextFormat *)rmessage.text;
		
		//if receives the movie relay start message, dump it on display window and monitor the stop key using the kbhit routines 
    	init_keyboard();
		if(rtext->SourceTaskNumber == CAMERA && rtext->code == START){ 
			while(1) {
				nc = read(tb[CAMERA].Pipe2Main, buf, BUFSIZ);
				if(nc > 0) {
					if(flag == 0) {
						fprintf(stderr, "\n*** Monitoring the moive relay service ****\n");
			 			flag = 1;
					}
					fprintf(stderr,"%s", buf);
				}
			//	if any key, send a signal to the camera task 
				if(kbhit()) {
					ch=readch();
					if(ch =='s'){
						fprintf(stderr,"\nThe movie relay service stops now\n");
						fprintf(flog,"Sent the user signal 1 to camera task\n"); 
						kill(tb[CAMERA].ProcessId, SIGUSR1); //send the user signal 1 to the camera task
						break;
					}
				}
			}
		}
    		close_keyboard();
    		init_keyboard();	//The following two lines are to fix temporary a problem that 
    		close_keyboard();	// that the console seldom displays the message of "Button: 0 ..." repeatly 

		while(1) {
			nc=read(tb[CAMERA].Pipe2Main,buf,sizeof(buf));
			//if receives the movie relay stop message from camera task, stop it 
			if(msgrcv(mqid, (void *)&rmessage, BUFSIZ, MAIN, 0) == -1) {
				fprintf(stderr, "Fail to receive the movie relay start message at main !!");
				exit(-1);
			}
			MessageLog(flog, "RX", rmessage);
			nrmessage++;
			rtext = (struct MovieTextFormat *)rmessage.text;
			if(rtext->SourceTaskNumber == CAMERA && rtext->code == FINISH)
				break;
		}
	}
}

void call()
{
	struct MessageFormat smessage;
	struct MSMSTextFormat mstext;
	char callNumber[50], voicefilename[50];
	int button, messagecount;
	
	printf("\nButton: 0(call),1(exit)\n");
	scanf("%d", &button);
	
	if(button == 0) {
		printf("Insert destination call number\n");
		printf("Insert : ");
		scanf("%s", &callNumber);
		printf("\n\nInsert what do you want many send.\n");
		printf("Insert : ");
		scanf("%d", &messagecount);
		printf("\n\nInsert voice file name.\n");
		printf("Insert : ");
		scanf("%s", &voicefilename);

		// send the movie relay request message to the camera task
		smessage.type = CALL;
		mstext.SourceTaskNumber = MAIN;
		mstext.code = REQUEST;
		strcpy(mstext.RecipientNumber, callNumber);
		strcpy(mstext.data, voicefilename);
		mstext.NumberOfMessages = messagecount;
		StructCopy(smessage.text, &mstext, BUFSIZ);
		// Send the prepared message
		if(msgsnd(mqid, (void *)&smessage, BUFSIZ, IPC_NOWAIT) == -1) {
			fprintf(stderr, "Fail to send message !!");
			exit(-1);
		}
		MessageLog(flog, "TX", smessage);
		nsmessage++;
	}
	else
		return;
}

void mail()
{
	struct MessageFormat smessage;
	struct MSMSTextFormat mstext;
	char destinationAddress[50], insert_message[200], c;
	int button, messagecount;
	
	printf("\nButton: 0(mail),1(exit)\n");
//	getchar(c);
//	button = atoi(&c);
	scanf("%d", &button);
	
	if(button == 0) {
		printf("nsert destination mail addressr\n");
		printf("Insert : ");
		//gets(destinationAddress);
		scanf("%s", &destinationAddress);
		printf("%s", destinationAddress);
		printf("\n\nInsert what do you want many send.\n");
		printf("Insert : ");
		scanf("%d", &messagecount);
		printf("%d", messagecount);
		printf("\n\nInsert message.\n");
		printf("Insert : ");
		//gets(message);
		scanf("%s", &insert_message);
		printf("\n%s\n", insert_message);
	
		printf("\ntest\n");

		// send the mass mail request message to the camera task
		smessage.type = WEB;
		mstext.SourceTaskNumber = MAIN;
		mstext.code = REQUEST;
		strcpy(mstext.RecipientName, destinationAddress);
		strcpy(mstext.data, insert_message);
		mstext.NumberOfMessages = messagecount;
		StructCopy(smessage.text, &mstext, BUFSIZ);
		// Send the prepared message
		if(msgsnd(mqid, (void *)&smessage, BUFSIZ, IPC_NOWAIT) == -1) {
			fprintf(stderr, "Fail to send message !!");
			exit(-1);
		}
		MessageLog(flog, "TX", smessage);
		nsmessage++;
	}
	else
		return;
}

void statistics()
{
	int i, button;

	printf("\nButton: 0(show statistics),1(exit)\n");
	scanf("%d", &button);
	
	if(button == 0) {
		printf("Diplay all shared data\n\n");
		for(i = 1; i < TASKS; i++) {
			printf("TASK[%d].ase : %d\n", i, board->ase[i]);
			printf("TASK[%d].ast : %d.%d\n", i, board->ast[i].tv_sec, board->ast[i].tv_usec);
		}
	}
}
			

//The following four functions of init_keyboard(), close_keyboard(), kbhit() and readch() 
// are to implement unblocking keyboard input function, 
// which is mandatory to stop the camera's movie realy during the console monitors the movie relay 

static struct termios initial_settings, new_settings;
static int peek_character = -1;

void init_keyboard()
{
    tcgetattr(0,&initial_settings);
    new_settings = initial_settings;
    new_settings.c_lflag &= ~ICANON;
    new_settings.c_lflag &= ~ECHO;
    new_settings.c_lflag &= ~ISIG;
    new_settings.c_cc[VMIN] = 1;
    new_settings.c_cc[VTIME] = 0;
    tcsetattr(0, TCSANOW, &new_settings);
}

void close_keyboard()
{
    tcsetattr(0, TCSANOW, &initial_settings);
}

int kbhit()
{
    char ch;
    int nread;

    if(peek_character != -1)
        return 1;
    new_settings.c_cc[VMIN]=0;
    tcsetattr(0, TCSANOW, &new_settings);
    nread = read(0,&ch,1);
    new_settings.c_cc[VMIN]=1;
    tcsetattr(0, TCSANOW, &new_settings);

    if(nread == 1) {
        peek_character = ch;
        return 1;
    }
    return 0;
}

int readch()
{
    char ch;

    if(peek_character != -1) {
        ch = peek_character;
        peek_character = -1;
        return ch;
    }
    read(0,&ch,1);
    return ch;
}
